# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod, abstractproperty
import os

from fuzzywuzzy import process
from mutagen.mp3 import EasyMP3

from .consts import MUSIC_LIBRARY_PATH
from .decorators import log_exectime
from .models import SongModel


providers = set()


class AbstractProvider(metaclass=ABCMeta):
    """abstract music resource provider"""

    @abstractproperty
    def name(self):
        """provider name, used to identify"""

    @abstractmethod
    def search(self, keyword, **kwargs):
        """search songs by name, artist name, album name

        :return: list of brief songs.
        """

    @abstractmethod
    def get_song(self, identifier):
        """get song by identifier

        :return: song model: :class:`fuocore.models.SongModel`
        """


class LocalProvider(AbstractProvider):

    def __init__(self, library_paths=[MUSIC_LIBRARY_PATH], depth=2):
        self._library_paths = library_paths
        self._library = list()
        self._songs = list()  # [(path, model)]
        self._identifier_song_map = dict()
        self.is_initialized = False

        # to initialize library
        self.library_paths = library_paths

    def _scan_dir(self, directory, exts=['mp3'], depth=1):
        if depth < 0:
            return []

        media_files = []
        for path in os.listdir(directory):
            path = os.path.join(directory, path)
            if os.path.isdir(path):
                files = self._scan_dir(path, exts, depth - 1)
                media_files.extend(files)
            elif os.path.isfile(path):
                if path.split('.')[-1] in exts:
                    media_files.append(path)
        return media_files

    @property
    def library_paths(self):
        return self._library_paths

    @library_paths.setter
    def library_paths(self, library_path):
        self._library = self.scan()
        self._songs = [self.model_from_fpath(fpath) for fpath in self._library]
        self._identifier_song_map.clear()
        if not self._identifier_song_map:
            for song in self._songs:
                self._identifier_song_map[song.identifier] = song

    @log_exectime
    def scan(self, exts=['mp3'], depth=1):
        """scan all media files in a directory"""
        depth = depth if depth <= 2 else 2
        media_files = []
        for directory in self._library_paths:
            media_files.extend(self._scan_dir(directory, exts, depth))
        return list(set(media_files))

    def add_to_library(self, directory):
        if directory not in self._library_paths:
            self._library_paths.append(directory)
        self.library_paths = self._library_paths

    @property
    def songs(self):
        return self._songs

    def model_from_fpath(self, fpath):
        metadata = EasyMP3(fpath)
        title = metadata['title'][0] if 'title' in metadata else 'unknown'
        artists_name = metadata['artist'] if 'artist' in metadata else ['unknown']
        brief_artists = [{'name': name} for name in artists_name]
        album_name = metadata['album'][0] if 'album' in metadata else 'unknown'
        brief_album = {'name': album_name}
        song = SongModel(name=title, brief_artists=brief_artists, url=fpath, brief_album=brief_album)
        song.source = self.name
        return song

    @property
    def name(self):
        return 'local'

    @log_exectime
    def search(self, keyword, limit=10):
        repr_song_map = dict()
        for song in self.songs:
            repr_song_map[str(song) + song.identifier] = song
        choices = repr_song_map.keys()
        result = process.extract(keyword, choices, limit=limit)
        result_songs = []
        for each, score in result:
            result_songs.append(repr_song_map[each])
        return result_songs

    def get_song(self, identifier):
        return self._identifier_song_map.get(identifier)


def register(provider):
    providers.add(provider)


def deregister(provider):
    if provider in providers:
        providers.remove(provider)



# -*- coding: utf-8 -*-

from .provider import register


class Source(object):

    def __init__(self):
        self._providers = set()

    def add_provider(self, provider):
        register(provider)
        self._providers.add(provider)

    def search(self, keyword):
        """search song by keyword

        :returns: list of :class:`fuocore.models.BriefSongModel`
        """
        songs = []
        for provider in self._providers:
            songs.extend(provider.search(keyword=keyword))
        return songs

    def get_song(self, identifier):
        """get song from identifier"""



# -*- coding: utf-8 -*-

import uuid

from april import Model
from april.tipes import listof
from april.utils import is_nested_type


class BaseModel(Model):

    _source = None

    @property
    def identifier(self):
        """the model indentifier (unique)"""
        if hasattr(self, '_identifier'):
            return self._identifier
        self._identifier = uuid.uuid1().hex
        return self._identifier

    @property
    def source(self):
        if self._source is None:
            raise NotImplementedError()
        return self._source

    @source.setter
    def source(self, value):
        """set source for model and model's field recursively"""
        self._source = value
        for name, ftype in self._fields:
            if issubclass(ftype, BaseModel):
                setattr(getattr(self, name), 'source', value)
            if is_nested_type(ftype):
                # FIXME: wait for ``april`` redesigning nested type interface
                for each in getattr(self, name):
                    setattr(each, 'source', value)

    def __eq__(self, other):
        if hasattr(other, 'identifier'):
            return self.identifier == other.identifier
        return False

    def __ne__(self, other):
        if hasattr(other, 'identifier'):
            return self != other
        return True


class BriefArtistModel(BaseModel):
    name = str
    img = str

    _optional_fields = ['img']


class BriefAlbumModel(BaseModel):
    name = str


class BriefSongModel(BaseModel):

    name = str
    url = str
    duration = int
    brief_album = BriefAlbumModel
    brief_artists = listof(BriefArtistModel)

    _optional_fields = ['url', 'duration', 'brief_album', 'brief_artists']

    def __str__(self):
        if not hasattr(self, 'brief_artists'):
            return self.name + ' - unknown'
        module = type(self).__module__
        class_name = type(self).__name__
        song_name = self.name + ' - ' + \
            ', '.join([artist.name for artist in self.brief_artists])
        source = self.source
        return '<{module}.{class_name} {song_name} {source}>'.format(**locals())


class ArtistModel(BriefArtistModel):
    songs = listof(BriefSongModel)
    desc = str

    _optional_fields = ('desc', )


class AlbumModel(BriefAlbumModel):
    img = str
    brief_songs = listof(BriefSongModel)
    brief_artists = listof(BriefArtistModel)
    desc = str

    _optional_fields = ('desc', )


class LyricModel(Model):
    song = BriefSongModel
    content = str
    trans_content = str

    _optional_fields = ('trans_content', )


class SongModel(BriefSongModel):

    _optional_fields = ['duration', 'brief_album', 'brief_artists',
                        'album', 'artists']



# -*- coding: utf-8 -*-
import logging

from .engine import AbstractPlayer, State, Playlist
from mpv import MPV, MpvEventID, MpvEventEndFile

logger = logging.getLogger(__name__)


class MpvPlayer(AbstractPlayer):
    """

    player will always play playlist current song. player will listening to
    playlist ``song_changed`` signal and change the current playback.
    """
    def __init__(self):
        super().__init__()
        self._mpv = MPV(ytdl=False,
                        input_default_bindings=True,
                        input_vo_keyboard=True)
        self._playlist = Playlist()
        self._playlist.song_changed.connect(self._on_song_changed)

    def initialize(self):
        self._mpv.observe_property(
            'time-pos',
            lambda name, position: self._on_position_changed(position)
        )
        self._mpv.observe_property(
            'duration',
            lambda name, duration: self._on_duration_changed(duration)
        )
        self._mpv.register_event_callback(lambda event: self._on_event(event))
        self.song_finished.connect(self._playlist.next)

    def quit(self):
        del self._mpv

    def play(self, url):
        logger.info('start play url:%s' % url)
        # clear playlist before play next song.
        # otherwise, mpv will seek to the last position and play
        self._mpv.playlist_clear()

        self._mpv.play(url)
        self.state = State.playing
        self.media_changed.emit()

    def play_song(self, song):
        if self.playlist.current_song is not None and \
                self.playlist.current_song == song:
            logger.warning('the song to be played is same as current song')
            return
        self._playlist.current_song = song

    def resume(self):
        self._mpv.pause = False
        self.state = State.playing

    def pause(self):
        self._mpv.pause = True
        self.state = State.paused

    def toggle(self):
        self._mpv.pause = not self._mpv.pause
        if self._mpv.pause:
            self.state = State.paused
        else:
            self.state = State.playing

    def stop(self):
        self._mpv.stop()
        self.state = State.stopped

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        self._mpv.seek(position, reference='absolute')
        self._position = position

    @AbstractPlayer.volume.setter
    def volume(self, value):
        super(MpvPlayer, MpvPlayer).volume.__set__(self, value)
        self._mpv.volume = self.volume

    def _on_position_changed(self, position):
        self._position = position
        self.position_changed.emit()

    def _on_duration_changed(self, duration):
        """listening to mpv duration change event"""
        logger.info('player receive duration changed signal')
        self.duration = duration

    def _on_song_changed(self):
        logger.info('player received song changed signal')
        if self._playlist.current_song is not None:
            logger.info('will play song: %s' % self._playlist.current_song)
            self.play(self._playlist.current_song.url)
        else:
            logger.info('playlist provide no song anymore.')

    def _on_event(self, event):
        if event['event_id'] == MpvEventID.END_FILE:
            reason = event['event']['reason']
            logger.info('current song finished. reason: %d' % reason)
            if reason != MpvEventEndFile.ABORTED:
                self.song_finished.emit()



#!/usr/bin/env python3

from setuptools import setup


requires = [
    'pycrypto>=2.6.1',
    'requests>=2.13.0',
    'beautifulsoup4>=4.5.3',
    'marshmallow>=2.13.5',
    'april==0.0.1a4',
    'aiozmq',
    'msgpack-python',
    'mutagen>=1.37',
    'python-Levenshtein>=0.12.0',
    'fuzzywuzzy',

]


setup(
    name='fuocore',
    version='0.0.6',
    description='feeluown core',
    author='Cosven',
    author_email='cosven.yin@gmail.com',
    py_modules=['mpv'],
    packages=[
        'fuocore',
        'fuocore.third_party',
        'fuocore.third_party.netease'
        ],
    package_data={
        '': []
        },
    url='https://github.com/cosven/feeluown-core',
    keywords=['media', 'player', 'api'],
    classifiers=(
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        ),
    install_requires=requires,
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
        'mock',
    ],
    entry_points={
        'console_scripts': []
        },
)
